/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.mezeo.storage;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.net.SocketTimeoutException;
import java.security.NoSuchAlgorithmException;
import java.util.Collection;
import java.util.Locale;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.encryption.Encryption;
import org.dasein.cloud.encryption.EncryptionException;
import org.dasein.cloud.mezeo.MezeoCloudStorage;
import org.dasein.cloud.storage.AbstractBlobStoreSupport;
import org.dasein.cloud.storage.CloudStoreObject;
import org.dasein.cloud.storage.FileTransfer;
import org.dasein.util.Jiterator;
import org.dasein.util.JiteratorPopulator;
import org.dasein.util.PopulatorThread;

import com.mezeo.credentials.BasicAuthCredentials;
import com.mezeo.exception.InvalidArgumentException;
import com.mezeo.exception.MezeoException;
import com.mezeo.exception.ParsingException;
import com.mezeo.exception.ResourceNotFoundException;
import com.mezeo.exception.ServerException;
import com.mezeo.exception.UnknownServerError;
import com.mezeo.generic.service.FileDownloader;
import com.mezeo.generic.service.Service;
import com.mezeo.resources.Account;
import com.mezeo.resources.Cloud;
import com.mezeo.resources.Container;
import com.mezeo.resources.Resource;
import com.mezeo.resources.ResourceType;
import com.mezeo.resources.factory.XML.ResourceFactoryXML;
import com.mezeo.resources.list.FileList;
import com.mezeo.util.ProgressInfo;
import com.sun.istack.internal.NotNull;

public class Blob extends AbstractBlobStoreSupport {

	private static final int RETRIES = 3;

	static private final Logger logger = Logger.getLogger(Blob.class);

	private MezeoCloudStorage provider;

	private Service service = null;

	private Cloud cloud;

	private Account account;

	public Blob(MezeoCloudStorage provider) {
		this.provider = provider;
	}

	@Override
	public void clear(String directory) throws CloudException, InternalException {

		removeDirectory(directory);
		// for (CloudStoreObject file : listFiles(directory)) {
		// if (file.isContainer()) {
		// removeDirectory(directory + "/" + file.getName());
		// } else {
		// removeFile(directory, file.getName(), false);
		// }
		// }

	}

	@Override
	public void removeDirectory(String directory) throws CloudException, InternalException {
		Container container = loadDirectory(directory);
		try {
			getService().deleteResource(container);
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}
	}

	private Account getAccount() throws InternalException {
		try {
			if (this.account == null) {
				this.account = getCloud(RETRIES).getAccount();
				getService().getResource(this.account);
			}
			return this.account;
		} catch (ServerException e) {
			e.printStackTrace();
			throw new InternalException();
		} catch (ParsingException e) {
			e.printStackTrace();
			throw new InternalException();
		} catch (InvalidArgumentException e) {
			e.printStackTrace();
			throw new InternalException();
		} catch (ResourceNotFoundException e) {
			e.printStackTrace();
			throw new InternalException();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
			throw new InternalException();
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException();
		}

	}

	private Service getService() throws UnsupportedEncodingException {
		if (this.service == null) {
			String userId = provider.getContext().getAccountNumber();
			String apiKey = new String(provider.getContext().getAccessPublic(), "utf-8");
			String password = new String(provider.getContext().getAccessPrivate(), "utf-8");
			BasicAuthCredentials auth = new BasicAuthCredentials(userId, password, apiKey);
			this.service = new Service(auth, userId);
		}
		return this.service;
	}

	@Override
	public String createDirectory(String bucket, boolean findFreeName) throws InternalException, CloudException {

		try {
			Service service = getService();
			Cloud cloud = getCloud(RETRIES);
			Container root = cloud.getRootContainer();

			service.getResource(root);

			createDirectory(root, bucket, findFreeName);

			return bucket;
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}
	}

	private void createDirectory(Container container, String bucket, boolean findFreeName) throws Exception {

		System.out.println("createDirectory(Container container, String bucket)- enter");

		String[] directories = bucket.split("/");
		for (int i = 0; i < directories.length; i++) {
			String directoryName = directories[i];

			Container bucketContainer = loadDirectory(container, directoryName);
			Container parentContainer = container;
			if (bucketContainer == null) {
				container = postContainer(parentContainer, directoryName, container.getResourceURI() + "/contents");
			} else {
				if (findFreeName && (i == directories.length - 1)) {
					container = postContainer(parentContainer, findFreeName(parentContainer, directoryName),
							container.getResourceURI() + "/contents");
				} else {
					container = bucketContainer;
				}
			}
		}

		System.out.println("createDirectory(Container container, String bucket)- exit");

	}

	private Container putContainer(Container parentContainer, String directoryName, String resourceURI)
			throws ServerException, ParsingException, InvalidArgumentException, ResourceNotFoundException, Exception {
		System.out.println("postContainer(Container parentContainer, String bucket, String resourceURI) - enter");
		System.out.println("posting: " + directoryName);
		System.out.println("resourceURI: " + resourceURI);
		Container container = (Container) service.createNewResource(ResourceType.CONTAINER_INFO);
		container.setName(directoryName);
		container.setResourceURI(resourceURI);
		if (parentContainer != null) {
			container.setParent(parentContainer);
		}
		service.putResource(container);
		System.out.println("postContainer(Container parentContainer, String bucket, String resourceURI) - exit");

		return container;
	}

	private Container postContainer(Container parentContainer, String bucket, String resourceURI) throws Exception,
			ServerException, ParsingException, InvalidArgumentException, ResourceNotFoundException {
		System.out.println("postContainer(Container parentContainer, String bucket, String resourceURI) - enter");
		System.out.println("posting: " + bucket);
		System.out.println("resourceURI: " + resourceURI);
		Container container = (Container) service.createNewResource(ResourceType.CONTAINER_INFO);
		container.setName(bucket);
		container.setParent(parentContainer);
		if (parentContainer != null) {
			container.setParent(parentContainer);
		}
		service.postResource(container, resourceURI);
		System.out.println("postContainer(Container parentContainer, String bucket, String resourceURI) - exit");

		return container;
	}

	protected String verifyName(String name, boolean container) throws CloudException {
		if (name == null) {
			return null;
		}
		StringBuilder str = new StringBuilder();
		name = name.toLowerCase().trim();

		if (name.length() > 255) {
			String extra = name.substring(255);
			int idx = extra.indexOf(".");

			if (idx > -1) {
				throw new CloudException("S3 names are limited to 255 characters.");
			}
			name = name.substring(0, 255);
		}
		while (name.indexOf("--") != -1) {
			name = name.replaceAll("--", "-");
		}
		while (name.indexOf("..") != -1) {
			name = name.replaceAll("\\.\\.", ".");
		}
		while (name.indexOf(".-") != -1) {
			name = name.replaceAll("\\.-", ".");
		}
		while (name.indexOf("-.") != -1) {
			name = name.replaceAll("-\\.", ".");
		}
		for (int i = 0; i < name.length(); i++) {
			char c = name.charAt(i);

			if (Character.isLetterOrDigit(c)) {
				str.append(c);
			} else {
				if (i > 0) {
					if (c == '/') {
						c = '.';
					} else if (c != '.' && c != '-') {
						c = '-';
					}
					str.append(c);
				}
			}
		}
		name = str.toString();
		while (name.indexOf("..") != -1) {
			name = name.replaceAll("\\.\\.", ".");
		}
		if (name.length() < 1) {
			return "000";
		}
		while (name.charAt(name.length() - 1) == '-' || name.charAt(name.length() - 1) == '.') {
			name = name.substring(0, name.length() - 1);
			if (name.length() < 1) {
				return "000";
			}
		}
		if (name.length() < 1) {
			return "000";
		} else if (name.length() == 1) {
			name = name + "00";
		} else if (name.length() == 2) {
			name = name + "0";
		}
		return name;
	}

	// @Override
	// public FileTransfer download(CloudStoreObject cloudFile, File diskFile)
	// throws CloudException, InternalException {
	//
	// String directory = cloudFile.getDirectory();
	// String fileName = cloudFile.getName();
	//
	// return download(directory, fileName, diskFile, null);
	// }

	@Override
	protected void get(@NotNull String directory, @NotNull String fileName, @NotNull File toFile,
			@NotNull FileTransfer transfer) throws IOException, CloudException {
		try {
			get(directory, fileName, toFile, transfer, null);
			if (transfer != null) {
				transfer.setPercentComplete(100D);
			}
		} catch (InternalException e) {
			e.printStackTrace();
			throw new IOException(e);
		}

	}

	protected void get(@NotNull final String directory, @NotNull final String fileName, @NotNull final File toFile,
			FileTransfer transfer, final Encryption encryption) throws IOException, CloudException, InternalException {
		try {

			final FileOutputStream outputStream;

			final File tempFile;
			try {
				if (encryption != null) {
					tempFile = File.createTempFile(fileName, ".tmp	");
					if (!tempFile.exists()) {
						tempFile.createNewFile();
					}
					outputStream = new FileOutputStream(tempFile);
				} else {
					tempFile = null;
					if (!toFile.exists()) {
						toFile.createNewFile();
					}

					outputStream = new FileOutputStream(toFile);
				}

				boolean success = false;

				try {
					com.mezeo.resources.File mezeoFile = loadFile(directory, fileName);
					if (mezeoFile != null) {
						FileDownloader downLoader = new FileDownloader(getService());
						ProgressInfo progressInfo = new ProgressInfo() {

							@Override
							public void setProgress(float percentage) {

							}
						};
						downLoader.downloadFile(mezeoFile, progressInfo, outputStream);

						if (encryption != null) {

							try {
								if (!toFile.exists()) {
									toFile.createNewFile();
								}
								FileInputStream input = new FileInputStream(tempFile);
								FileOutputStream output;

								output = new FileOutputStream(toFile);
								encryption.decrypt(input, output);
								output.flush();
								output.close();
								input.close();

							} catch (EncryptionException e) {
								e.printStackTrace();
								throw new InternalException(e);
							} catch (IOException e) {
								e.printStackTrace();
								throw new InternalException(e);
							}

						}

						success = true;
					} else {
						success = false;
					}

				} catch (MezeoException e) {
					e.printStackTrace();
					throw new InternalException(e);
				} finally {
					if (!success) {
						if (toFile.exists()) {
							toFile.delete();
						}
					}
					if (tempFile != null) {
						tempFile.delete();
					}

				}

			} catch (FileNotFoundException e) {
				e.printStackTrace();
				throw new InternalException(e);
			} catch (IOException e) {
				e.printStackTrace();
				throw new InternalException(e);
			}

		} finally {
			System.out
					.println("download(final String directory, final String fileName, final File toFile, final Encryption encryption) - exit");
		}
	}

	@Override
	public boolean exists(String bucket) throws InternalException, CloudException {
		try {
			Cloud cloud = getCloud(RETRIES);
			Container container = cloud.getRootContainer();

			if (bucket != null) {
				String[] directories = bucket.split("/");
				for (String directoryName : directories) {
					container = loadDirectory(container, directoryName);
					if (container == null) {
						return false;
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}
		return true;
	}

	@Override
	public long exists(String directory, String fileName, boolean multiPart) throws InternalException, CloudException {

		com.mezeo.resources.File file = loadFile(directory, fileName);
		if (file != null) {
			return file.getSize();
		}
		return -1;

	}

	public boolean exists(Container container, String directoryName) throws InternalException {
		try {
			Container bucketContainer = loadDirectory(container, directoryName);
			if (bucketContainer == null) {
				return false;
			}
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}
		return true;

	}

	@Override
	public long getMaxFileSizeInBytes() throws InternalException {
		// max file size = what's available on the server.
		return getAccount().getAllocatedStorage() - getAccount().getUsedStorage();
	}

	@Override
	public String getProviderTermForDirectory(Locale arg0) {
		return "container";
	}

	@Override
	public String getProviderTermForFile(Locale arg0) {
		return "file";
	}

	@Override
	public boolean isPublic(String directory, String file) throws CloudException, InternalException {
		try {
			com.mezeo.resources.File mezeoFile = loadFile(directory, file);
			if (mezeoFile != null) {
				return mezeoFile.isPublicFile();
			}
			return false;
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}
	}

	@Override
	public Collection<CloudStoreObject> listFiles(String bucket) throws CloudException, InternalException {
		try {
			PopulatorThread<CloudStoreObject> populator;

			Cloud cloud = getCloud(RETRIES);
			Container container = cloud.getRootContainer();

			String finalDirectoryName = null;
			Container finalContainer = null;
			if (bucket != null) {
				String[] directories = bucket.split("/");
				for (String directoryName : directories) {
					container = loadDirectory(container, directoryName);
					finalDirectoryName = directoryName;
					finalContainer = container;
				}
			}

			final String dir = finalDirectoryName;
			final Container cont = finalContainer;
			populator = new PopulatorThread<CloudStoreObject>(new JiteratorPopulator<CloudStoreObject>() {
				public void populate(Jiterator<CloudStoreObject> iterator) throws CloudException, InternalException {
					listFiles(cont, dir, iterator);
				}
			});
			populator.populate();
			return populator.getResult();
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}
	}

	private void listFiles(Container container, String bucket, Jiterator<CloudStoreObject> iterator)
			throws CloudException, InternalException {
		loadResources(container, bucket, iterator, LoadingResourceFlag.FILES_AND_DIRECTORIES);
	}

	private Container loadDirectory(String directory) throws InternalException {

		Container container;
		try {
			Cloud cloud = getCloud(RETRIES);
			container = cloud.getRootContainer();

			String[] directories = directory.split("/");
			for (int i = 0; i < directories.length; i++) {
				String pathSegment = directories[i];
				container = loadDirectory(container, pathSegment);
			}
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}

		return container;
	}

	private com.mezeo.resources.File loadFile(String directory, String fileName) throws InternalException {
		try {
			Service service = getService();
			Container container;
			if (directory == null) {
				Cloud cloud = getCloud(RETRIES);
				container = cloud.getRootContainer();
				service.getResource(container);
			} else {
				container = loadDirectory(directory);
			}

			if (container != null) {
				String resourceURI = container.getResourceURI() + "/contents";

				FileList fileList = (FileList) ResourceFactoryXML.createNewResource(ResourceType.FILE_LIST);
				fileList.addHeader(Service.X_CLOUD_DEPTH_STR, "1");
				container.setFileContents(fileList);

				fileList.setResourceURI(resourceURI);
				service.getResource(fileList);

				Vector<Resource> victor = fileList.getResourceList();
				for (Resource resource : victor) {
					service.getResource(resource);
					if (resource instanceof com.mezeo.resources.File
							&& ((com.mezeo.resources.File) resource).getName().equals(fileName)) {
						((com.mezeo.resources.File) resource).setParent(container);
						return (com.mezeo.resources.File) resource;
					}
				}
			}
			return null;
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException();
		}
	}

	private Container loadDirectory(Container container, String directory) throws ServerException, ParsingException,
			InvalidArgumentException, ResourceNotFoundException, Exception {
		Service service = getService();

		service.getResource(container);
		String resourceURI = container.getResourceURI() + "/contents";

		FileList fileList = (FileList) ResourceFactoryXML.createNewResource(ResourceType.FILE_LIST);
		fileList.addHeader(Service.X_CLOUD_DEPTH_STR, "1");
		container.setFileContents(fileList);

		fileList.setResourceURI(resourceURI);
		service.getResource(fileList);

		Container soughtContainer = null;
		Vector<Resource> victor = fileList.getResourceList();
		for (Resource resource : victor) {
			service.getResource(resource);
			if (resource instanceof Container && ((Container) resource).getName().equals(directory)) {
				soughtContainer = (Container) resource;
				break;
			}

		}
		return soughtContainer;
	}

	public enum LoadingResourceFlag {
		FILES, DIRECTORIES, FILES_AND_DIRECTORIES;

		public boolean isFiles() {
			return this.equals(FILES) || this.equals(FILES_AND_DIRECTORIES);
		}

		public boolean isDirectories() {
			return this.equals(DIRECTORIES) || this.equals(FILES_AND_DIRECTORIES);
		}

	}

	private void loadResources(Container container, String bucket, Jiterator<CloudStoreObject> iterator,
			LoadingResourceFlag flag) throws CloudException, InternalException {
		Service service;

		try {
			service = getService();
		} catch (UnsupportedEncodingException e) {
			throw new InternalException(e);
		}
		try {

			if (container == null) {
				Cloud cloud = getCloud(RETRIES);
				container = cloud.getRootContainer();
				service.getResource(container);
			}

			String resourceURI = container.getResourceURI() + "/contents";

			FileList fileList = (FileList) ResourceFactoryXML.createNewResource(ResourceType.FILE_LIST);
			fileList.addHeader(Service.X_CLOUD_DEPTH_STR, "1");
			container.setFileContents(fileList);

			fileList.setResourceURI(resourceURI);
			service.getResource(fileList);

			Vector<Resource> victor = fileList.getResourceList();
			for (Resource resource : victor) {
				service.getResource(resource);

				CloudStoreObject file = new CloudStoreObject();

				if (flag.isFiles() && (resource instanceof com.mezeo.resources.File)) {
					file.setContainer(false);
					file.setName(((com.mezeo.resources.File) resource).getName());
				}
				if (flag.isDirectories() && resource instanceof Container) {
					file.setContainer(true);
					file.setName(((Container) resource).getName());
				}

				iterator.push(file);
			}

		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}

	}

	private Cloud getCloud(int tries) throws InternalException, CloudException {
		if (this.cloud == null) {
			if (tries == 0) {
				throw new CloudException("Number of tries exceeded getting cloud.");
			}
			Cloud cld = (Cloud) ResourceFactoryXML.createNewResource(ResourceType.CLOUD_INFO);
			cld.setResourceURI(provider.getContext().getEndpoint());
			cld.addHeader(Service.X_CLOUD_DEPTH_STR, "1");

			Container rootContainer = (Container) ResourceFactoryXML.createNewResource(ResourceType.CONTAINER_INFO);
			rootContainer.addHeader(Service.X_CLOUD_DEPTH_STR, "1");
			cld.setRootContainer(rootContainer);

			try {
				getService().getResource(cld);
			} catch (SocketTimeoutException e) {
				logger.warn("SocketTimeoutException getting cloud. Retrying: " + (tries - 1) + " tries left.");
				System.out.println("Socket Timeout Exception getting cloud. Retrying: " + (tries - 1) + " tries left.");
				getCloud(tries - 1);
			} catch (UnknownServerError e) {
				logger.warn("UnknownServerError getting cloud. Retrying: " + (tries - 1) + " tries left.");
				System.out.println("UnknownServerError getting cloud. Retrying: " + (tries - 1) + " tries left.");
				getCloud(tries - 1);
			} catch (Exception e) {

				e.printStackTrace();
				throw new InternalException(e);
			}
			this.cloud = cld;
		}
		return this.cloud;
	}

	@Override
	public void makePublic(String directory) throws InternalException, CloudException {
		try {
			Container container = loadDirectory(directory);
			container.setShared(true);

			service.putResource(container);
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}
	}

	@Override
	public void makePublic(String directory, String fileName) throws InternalException, CloudException {
		try {
			com.mezeo.resources.File mezeoFile = loadFile(directory, fileName);
			if (mezeoFile != null) {
				mezeoFile.setPublicFile(true);

				service.putResource(mezeoFile);
			}
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}

	}

	@Override
	public void moveFile(String sourceContainer, String file, String targetContainer) throws InternalException,
			CloudException {
		try {
			com.mezeo.resources.File mezeoFile = loadFile(sourceContainer, file);
			Container parent = loadDirectory(targetContainer);
			mezeoFile.setParent(parent);

			service.postResource(mezeoFile, mezeoFile.getResourceURI());
			removeFile(sourceContainer, file, true);

		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}

	}

	@Override
	public String renameDirectory(String oldName, String newName, boolean findFreeName) throws CloudException,
			InternalException {
		Container container = loadDirectory(oldName);
		try {
			if (findFreeName) {
				findFreeName(container.getParent(), newName);
			}

			service.postResource(container, container.getResourceURI());
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}
		return container.getName();
	}

	@Override
	public void renameFile(String directory, String file, String newName) throws CloudException, InternalException {
		try {
			com.mezeo.resources.File mezeoFile = loadFile(directory, file);
			mezeoFile.setName(newName);

			service.postResource(mezeoFile, mezeoFile.getResourceURI());
		} catch (Exception e) {
			e.printStackTrace();
			throw new InternalException(e);
		}
	}

	@Override
	public void upload(File source, String directory, String fileName, boolean multipart, Encryption encryption)
			throws CloudException, InternalException {
		if (!exists(directory)) {
			createDirectory(directory, false);
		}
		if (encryption != null) {
			multipart = true;
		}
		if (multipart) {
			try {
				uploadMultipartFile(source, directory, fileName, encryption);
			} catch (InterruptedException e) {
				logger.error(e);
				e.printStackTrace();
				throw new CloudException(e.getMessage());
			}
		} else {
			try {
				put(directory, fileName, source);
			} catch (IOException e) {
				throw new InternalException(e);
			}
		}
	}

	private String findFreeName(Container container, String bucket) throws InternalException, CloudException {
		int idx = bucket.lastIndexOf(".");
		String prefix, rawName;

		if (idx == -1) {
			prefix = null;
			rawName = bucket;
			bucket = rawName;
		} else {
			prefix = bucket.substring(0, idx);
			rawName = bucket.substring(idx + 1);
			bucket = prefix + "." + rawName;
		}
		while (exists(container, bucket)) {
			idx = rawName.lastIndexOf("-");
			if (idx == -1) {
				rawName = rawName + "-1";
			} else if (idx == rawName.length() - 1) {
				rawName = rawName + "1";
			} else {
				String postfix = rawName.substring(idx + 1);
				int x;

				try {
					x = Integer.parseInt(postfix) + 1;
					rawName = rawName.substring(0, idx) + "-" + x;
				} catch (NumberFormatException e) {
					rawName = rawName + "-1";
				}
			}
			if (prefix == null) {
				bucket = rawName;
			} else {
				bucket = prefix + "." + rawName;
			}
		}
		return bucket;
	}

	@Override
	protected void put(String directory, String fileName, File source) throws IOException, CloudException {
		try {
			Container container = loadDirectory(directory);

			Service service = getService();
			com.mezeo.resources.File file = service.multipartUpload(container.getResourceURI() + "/contents",
					source.getAbsolutePath());

			file.setName(fileName);
			file.setParent(container);
			if (!service.putResource(file)) {
				throw new CloudException(fileName + " did not successfully upload to directory: " + directory);
			}
			for (int tries = 0; exists(directory, fileName, false) < 0 && tries < 30; tries++) {
				try {
					Thread.sleep(1000L);
				} catch (InterruptedException e) {
				}
			}

		} catch (ServerException e) {
			throw new CloudException(e);
		} catch (ParsingException e) {
			throw new IOException(e);
		} catch (InvalidArgumentException e) {
			throw new IOException(e);
		} catch (ResourceNotFoundException e) {
			throw new IOException(e);
		} catch (InternalException e) {
			throw new IOException(e);
		} catch (Exception mezeoShouldNotThrowGenericExceptionsLikeThisShameOnTheirDevelopers) {
			throw new IOException(mezeoShouldNotThrowGenericExceptionsLikeThisShameOnTheirDevelopers);
		}
	}

	@Override
	protected void put(String directory, String fileName, String content) throws NoSuchAlgorithmException, IOException,
			CloudException {

		File file = null;
		try {
			file = File.createTempFile(fileName, ".txt");

			PrintWriter writer = new PrintWriter(new OutputStreamWriter(new FileOutputStream(file)));
			writer.print(content);
			writer.flush();
			writer.close();

			put(directory, fileName, file);

		} finally {
			if (file != null) {
				file.delete();
			}
		}
	}

	@Override
	protected void removeFile(String directory, String name) throws CloudException, InternalException {

		com.mezeo.resources.File file = null;
		for (int i = 5; i > 0 && file == null; i--) {
			file = loadFile(directory, name);
		}
		if (file != null) {
			try {
				getService().deleteResource(file);
			} catch (Exception e) {
				e.printStackTrace();
				throw new InternalException(e);
			}
		}
	}

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        return true;
    }
}
